/*   Main file of the galaxy simulation.
     The simulation principle is HOHL article's (1971) principle.
     FFT are not used becasue of poor precision on windows 32 bits.
     Therefore the calculation principle is based on HOHL's article idea and is
     encoded in what is called CalculatePotentialCell3() function in particle.c.    
     (This code has been used also for calculating the bullet cluster, in another version).
*/

/*************************************************************************************************************/
/*************************************************************************************************************/

#include <stdlib.h> 
#include <string.h> 
#include <stdio.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif
#include "common.h"
#include "init.h"
#include "particles.h"
#include "potential.h"
#include "hmi.h"
#include "stat.h"
#include "stet.h"

unsigned long int StepNum;   // StepNum. Here for debug and tracking precision and time.
int totaltime;               // For HMI only.

int main(int argc,char *argv[]) 
{
  int isConfigRead;          // NOT 0 if init is read partially from a file.
	fprintfPerso(stdout, " ----------------- STARTED -------------- \n");

  /******* Initialisation of the modules ***********************************/
  InitParticles();                // Init of the grid, the arrays...
  InitPotential();                // Init of the potential calculations.
  InitInit();                     // Init of the galaxy shape.
  InitStet();                     // Init STET mechanism if needed.
  isConfigRead = TryToReadConfig();   // Possible re-init of "init" and "stet". 
  InitHmi();                      // Init of HMI.
  StepNum = GetStepNum(isConfigRead); // Get StepNum from file or set it to 0.       

  /******* First phase of the simulation : going to some permanent regime **/
	for( ; StepNum < NB_STEP_FIRSTPHASE ; StepNum++)
	{
    /************* Saving config **********************************************/
    PlotConfig();                     // Will be used by InitGalaxy() later on.
    /******* Plotting ******************************************************/
		PlotOpenCloseFileIfNeeded(StepNum);    // Open or close files of streams.
		if(StepNum % PRINT_ECHO == 0) PrintSimulInfoInFile(stdout,0);
		PlotPartsOneStep(StepNum);   // Print posold for each particle.

    /******* Specific action if solar system configuration *****************/
		RazSolSyst();                // If needed, forcing partic simulating 
                                 // sun to be fixed on same location.
    /******* Recenter the galaxy in the middle of the grid *****************/
    #if PERIODIC_CALCUL == 0     // if "normal case".
      RecenterGrid();            // Put galactic center in grid middle. 
    #endif                       // 

    /******* Calculate new gravitational field *******************/
    DensityCalculate();          // Counting nb of part in each cell.
    ReInitStet();                // Reinit of StetCorrMax, StetCorrMin.
		GravitPotentialCalculate();  // Calculate gravitational potential.

    /******* Calculate new particles's position *****************************/
    CalcStetAddedNbPart();       // Progressively passing from Newton to STET.
    ProcessPosit();              // Calculation of positioning the particles.

    /************* Statistics **********************************************/
    #if PERIODIC_CALCUL == 0
      StatCalc();                // Report speed dispersion
    #endif
	}

  /************** Cooling actual state of the galaxy ***********************/
  Cooling();       // DEPRECATED This is never executed in the final version.

  /******* DEPRECATED. Simulation second phase: from COOLED permanent regime */
  /* This is never executed in the final's article  version DEPRECATED.      */
	for(StepNum = NB_STEP_FIRSTPHASE ; StepNum < NB_STEP ; StepNum++)
	{
    /******* Plotting ********************************************/
		PlotOpenCloseFileIfNeeded(StepNum);
		if(StepNum % PRINT_ECHO == 0) PrintSimulInfoInFile(stdout,0);
		PlotPartsOneStep(StepNum);   // Print posold for each particle.

    /******* Specific action if solar system configuration *****************/
		RazSolSyst();                // If needed, forcing partic simulating 
                                 // sun to be fixed on same location.
    /******* Calculate new gravitational field *******************/
    #if PERIODIC_CALCUL == 0
      RecenterGrid();
    #endif
    DensityCalculate();           // Counting nb of part in each cell.
		GravitPotentialCalculate();   // Calculate gravitational potential.

    /******* Calculate new particles's position ******************/
		ProcessPosit();              // Calculation of positioning the particles.

		total_time = CalculTime(total_time);   // Add dt to time. For plotting.

    /************* Statistics **********************************************/
    #if PERIODIC_CALCUL == 0
      StatCalc();           // Report speed dispersion
    #endif
	}

	PlotEndFile();        
	fprintfPerso(stdout, " ----------------- END -------------- \n");
	return 0;
}

unsigned long int GetStepNum(int isReadConfig)        
{
  if(isReadConfig == 0)
  {                       // Empty file : init from scratch (InitPart).
    return(0);            // And start from 0.
  }
  else return(StepNum);   // Here StepNum init has been done in TryToReadConfig.
}
/********************* Error function *************************/
void Error(int ErrNum)
{
  int WaitBuff;
  fprintfPerso(stdout, "ERROR, ErrNum=%d\n", ErrNum); // Print error number.
  scanf("%d",&WaitBuff);                              // Stop until ack by user.
}

double CalculTime(double total_time)
{
	return(total_time + Gravit_dt);
}
