/*   HANDLING PARTICLES (stars in a "galaxy").
*/

#include <stdlib.h> 
#include <string.h> 
#include <stdio.h>
#include <math.h>
#ifdef _OPENMP          
#include <omp.h>
#endif
#include "common.h"
#include "hmi.h"
#include "init.h"
#include "particles.h"
#include "stet.h"
#include "potential.h"
#include "fft.h"

/* ---------------------------- Engine variables --------------------------- */
particle_t Part[NBMAX_PART];
double Gravit_dt;           // Time slice copied in this glob var for debug, hmi prints.
unsigned int NbPart;        // Now calculated dynamically because of complicated init.
double total_time;
double PartMass;
unsigned int NbPartOut;        // Out of calculation grid.
unsigned int CellNbPart[NBCELLX_FFT][NBCELLY_FFT][NBCELLZ_FFT];
double GalCenter[3];

/* ------------------ Prototypes of internal functions --------------------- */
void GetAccelFromGalaxy(int PartInd, double accel[], unsigned int NumCell[]);

/* ------------------------ Init of the grid ------------------------------- */
void InitParticles()
{
  NbPartOut = 0;
  memset(CellNbPart, 0, sizeof(CellNbPart));
  VectorZero(GalCenter);
}

/* -------------------- Get the cell containing a given particle ----------- */
unsigned char GetCellNonPeriod(double pos[], unsigned int NumCell[])
{
  unsigned char ReturnCode;
  double CellOffset[3];

  /* --------------------- Calculate NumCell if IN ------------------------- */
  NumCell[2] = 0;                                             // Default value.      
  CellOffset[0] = pos[0] - GRID_ORIGIN0;    
  if(CellOffset[0] < 0  ||  CellOffset[0] > LG_GRID)
  {
    ReturnCode = CELL_OUTX;
  }
  else
  {
    CellOffset[1] = pos[1] - GRID_ORIGIN1;    
    if(CellOffset[1] < 0  ||  CellOffset[1] > LG_GRID)
    {
      ReturnCode = CELL_OUTY;
    }
    else
    {
      NumCell[0] = (unsigned int) (CellOffset[0] / ((double) LGX_CELL));
      NumCell[1] = (unsigned int) (CellOffset[1] / ((double) LGY_CELL));
      #if (NB_DIM == 3)
        CellOffset[2] = pos[2] - GRID_ORIGIN2;    
        if(CellOffset[2] < 0  ||  CellOffset[2] > LG_GRID)
        {
          ReturnCode = CELL_OUTZ;
        }

        else
        {
          NumCell[2] = (unsigned int) (CellOffset[2] / ((double) LGZ_CELL));
      #else
          NumCell[2] = 0;                 
      #endif
          if(GridDisk[NumCell[0]][NumCell[1]][NumCell[2]])
          {
            ReturnCode = CELL_IN;
          }
          else
          {
            ReturnCode = CELL_OUTX;      // Change return code later on.
          }
      #if (NB_DIM == 3)
        }
      #endif
    }
  }

  return(ReturnCode);
}

unsigned char GetCell(double pos[], unsigned int NumCell[])
{
  unsigned char ReturnCode;

  #if PERIODIC_CALCUL == 1
    ReturnCode = GetCellPeriod(pos, NumCell);
  #else
    ReturnCode = GetCellNonPeriod(pos, NumCell);
  #endif

  return(ReturnCode); 
}

/* ------- Calculate number of particles in each cells of the grid --------- */
void DensityCalculate()
{
  unsigned int i;
  unsigned CR;
  unsigned int NumCell[3];

  NumCell[2] = 0;
  NbPartOut = 0;
  memset(CellNbPart, 0, sizeof(CellNbPart));

  for(i = 0; i < NbPart; i++)
  {
    CR = GetCell(Part[i].pos, NumCell);
    switch(CR)
    {
      case CELL_OUTX:
      case CELL_OUTY:
      case CELL_OUTZ:
        NbPartOut++;
        break;
      case CELL_IN:
	      if(NumCell[0] < 0) Error(100);
	      if(NumCell[1] < 0) Error(101);
	      if(NumCell[2] < 0) Error(102);
	      if(NumCell[0] >= NBCELLX_GRID) Error(103);
	      if(NumCell[1] >= NBCELLY_GRID) Error(104);
	      if(NumCell[2] >= NBCELLZ_GRID) Error(105);
        CellNbPart[NumCell[0]][NumCell[1]][NumCell[2]]++;
        break;
      default:
        Error(106);
    }
	}
}

void GravitPotentialCalculateByGrid()
{
  int ix;                   // Signed because of parallel need.

  /* ---------------- Parallelisation on multi-core --------------------- */
  //#ifdef _OPENMP          // Noticeably avoid parallelisation with STEP by STEP.
    #pragma omp parallel for private( ix )     // Warning doesnt works of course
  //#endif                                       // with step by step. 

  /* ---------------- For each cell in the grid calculate its potential - */
  for(ix=0 ; ix<NBCELLX_GRID ; ix++)
  {
    unsigned int iy, iz;
    for(iy=0 ; iy<NBCELLY_GRID ; iy++)
    {
      for(iz=0 ; iz<NBCELLZ_GRID ; iz++)
      {
          switch(MODE_CALCUL_POTENTIAL)
          {
            case MODE_GRID:  // Calculate the integral convolution step by step.
              CalculatePotentialCell(ix, iy, iz); 
              break;
            case MODE_GRID3: // Calcul integral convol step by step optimized manner.
              CalculatePotentialCell3(ix, iy, iz);
              break;
          }
      }
    }
  }
}

void GravitPotentialCalculate()
{
  switch(MODE_CALCUL_POTENTIAL)
  {
    case MODE_FFT:          // Calculate the convolution by product in freq space.
      GravitPotentialCalculateByFFT();
      break;
    case MODE_GRID:          // Calculate the integral convolution step by step.
    case MODE_GRID3:         // Calcul integral convol step by step optimized manner.
      GravitPotentialCalculateByGrid();
      break;
  }
}

/* ----- Calculate next position of a particle which is far from galactic center ------- */
void ProcessPositOut(int PartInd, unsigned int NumCell[])      
{
  double dt2;

  /********* Calculate acceleration from the galaxy ******************/
  GetAccelFromGalaxy(PartInd, Part[PartInd].acc, NumCell);
      
  /**** Calculate next position from previously calculated accel *****/
  dt2 = Gravit_dt * Gravit_dt;  
  Part[PartInd].posnew[0] =  - Part[PartInd].posold[0] + 2 * Part[PartInd].pos[0]
      			                                        + Part[PartInd].acc[0] * dt2;
  Part[PartInd].posnew[1] =  - Part[PartInd].posold[1] + 2 * Part[PartInd].pos[1]
      			                                        + Part[PartInd].acc[1] * dt2;
  #if (NB_DIM == 3)
    Part[PartInd].posnew[2] =  - Part[PartInd].posold[2] + 2 * Part[PartInd].pos[2]
      			                                        + Part[PartInd].acc[2] * dt2;
  #endif
}

/* ----- Calculate next posit of a partic inside the grid ------------- */
void ProcessPositIn(int PartInd, unsigned int NumCell[])      
{
  double dt2;

  /**** Calculate acceleration from previously calculated potential *****/
  #if (NB_DIM == 3)
    if(     NumCell[0]<=0  ||  NumCell[1]<=0  ||  NumCell[2]<=0
        ||  NumCell[0]>=NBCELLX_GRID-1  ||  NumCell[1]>=NBCELLY_GRID-1 
                                        ||  NumCell[2]>=NBCELLZ_GRID-1 )
  #else
    if( NumCell[0]<=0  ||  NumCell[1]<=0  ||  NumCell[0]>=NBCELLX_GRID-1
                                        ||  NumCell[1]>=NBCELLY_GRID-1 )
  #endif
  {    // On the fringe. Impossible to calculate acceleration in a coherent way.
    Part[PartInd].acc[0] = Part[PartInd].acc[1] = 0; // Not perfect, should be avoided.
    #if (NB_DIM == 3)
      Part[PartInd].acc[2] = 0;
    #endif
  }
  else
  {                                /* Done in a "time reversal" manner */
      Part[PartInd].acc[0] =  Potential[NumCell[0]+1][NumCell[1]][NumCell[2]];
      Part[PartInd].acc[0] -= Potential[NumCell[0]-1][NumCell[1]][NumCell[2]];
      Part[PartInd].acc[0] /= 2 * LGX_CELL;
      Part[PartInd].acc[1] =  Potential[NumCell[0]][NumCell[1]+1][NumCell[2]];
      Part[PartInd].acc[1] -= Potential[NumCell[0]][NumCell[1]-1][NumCell[2]];
      Part[PartInd].acc[1] /= 2 * LGY_CELL;
    #if (NB_DIM == 3)
      Part[PartInd].acc[2] =  Potential[NumCell[0]][NumCell[1]][NumCell[2]+1];
      Part[PartInd].acc[2] -= Potential[NumCell[0]][NumCell[1]][NumCell[2]-1];
      Part[PartInd].acc[2] /= 2 * LGZ_CELL;
    #endif
  }
  
  /**** Calculate next position from previously calculated accel *******/
  dt2 = Gravit_dt * Gravit_dt;  
  Part[PartInd].posnew[0] = - Part[PartInd].posold[0] + 2 * Part[PartInd].pos[0]
      			                                      + Part[PartInd].acc[0] * dt2;
  Part[PartInd].posnew[1] = - Part[PartInd].posold[1] + 2 * Part[PartInd].pos[1]
      			                                      + Part[PartInd].acc[1] * dt2;
  #if (NB_DIM == 3)
    Part[PartInd].posnew[2] = - Part[PartInd].posold[2] + 2 * Part[PartInd].pos[2]
      			                                         + Part[PartInd].acc[2] * dt2;
  #endif

  MemorizeStetCorr(PartInd, NumCell);
}

/* ------------------ Calculate next position of a particle ----------------- */
void ProcessPosit()      
{
  unsigned int i;
  unsigned CR;
  unsigned int NumCell[3];
  double L = 1.E18;

  for(i = 0; i < NbPart; i++)
  {
    CR = GetCell(Part[i].pos, NumCell);
    switch(CR)
    {
      case CELL_OUTX:
      case CELL_OUTY:
      case CELL_OUTZ:
        ProcessPositOut(i, NumCell);      
        break;
      case CELL_IN:
        ProcessPositIn(i, NumCell);      
        break;
      default:
        Error(107);
    }
  }                 // End of for loop over the i particles.

  /* ----- Now that new positionS are calculated, update old and news ---------- */
  for(i = 0; i < NbPart; i++)
  {
    VectorCopy(Part[i].pos, Part[i].posold);
    VectorCopy(Part[i].posnew, Part[i].pos);
  }
}

/* -------------- Calculate acceleration coming the whole galaxy -------------- */
void GetAccelFromGalaxy(int PartInd, double accel[], unsigned int NumCell[])
{
  long double SquareDistance;
  double dx0, dx1, dx;
  #if (NB_DIM == 3)
    double dx2;
  #endif
  double acc;                      // Temporary variable memorizing acceleration.
  double GalaxyMass;

  /* ----- get distance^2 from galactic center ----- */
  dx0 = 0 - Part[PartInd].pos[0];
  dx1 = 0 - Part[PartInd].pos[1];
  SquareDistance = dx0*dx0 + dx1*dx1;
  #if (NB_DIM == 3)
    dx2 = 0 - Part[PartInd].pos[2];
    SquareDistance += dx2*dx2;
  #endif
  dx = sqrt(SquareDistance);

  /* ----- Retrieve cell mass ------------------ */
  GalaxyMass = (NbPart - NbPartOut) * PartMass;

  /* ------- force = G * p1_mass * pd2->mass / SquareDistance ---- */
  if (SquareDistance != 0) acc = GalaxyMass * G_CONST / SquareDistance;
  else acc = 0;                     // Rare case: avoiding any effect.
  accel[0] = (dx0/dx) * acc;        // x composant of accel vector.
  accel[1] = (dx1/dx) * acc;        // y composant of accel vector.
  #if (NB_DIM == 3)
    accel[2] = (dx2/dx) * acc;      // z composant of accel vector.
  #endif

  GetAccelFromGalaxyStet(PartInd, accel, NumCell);
}

/* ---------------- Getting the galactic center from max NbPart ------------ */
unsigned int GetOriginFromMaxNbPart(double PosOr[])
{
  int ix, iy, iz, jx, jy, jz, MaxNbP, NbP;

  iz = jz = 0;                      // 2D only here.
  PosOr[2] = 0;                     // 2D only here.
  MaxNbP = 0;
  for(ix=0 ; ix<NBCELLX_GRID ; ix++)
  {
    for(iy=0 ; iy<NBCELLY_GRID ; iy++)
    {
      NbP = 0;
      for(jx=-MATDENS_MEAN_RAY ; jx<=MATDENS_MEAN_RAY ; jx++)
      {
        for(jy=-MATDENS_MEAN_RAY ; jy<=MATDENS_MEAN_RAY ; jy++)
        {
          if(0<=ix+jx && ix+jx<NBCELLX_GRID && 0<=iy+jy && iy+jy<NBCELLY_GRID)
          {
            NbP += CellNbPart[ix+jx][iy+jy][iz+jz];
          }
        }
      }

      if( NbP > MaxNbP)
      {
        MaxNbP = NbP;
        PosOr[0] = GRID_ORIGIN0 + ix * LGX_CELL;
        PosOr[1] = GRID_ORIGIN1 + iy * LGY_CELL;
      }
    }
  }
  return(MaxNbP);
}

/* ---------------- Repositionning the grid "around" the galaxy ------------ */
void RecenterGrid()
{
  unsigned int i;
  double OffsetOrigin[3];
  
  GetOriginFromMaxNbPart(GalCenter);

  /* ------ Add a manual offset ------------------------------------------ */
  OffsetOrigin[0] = OFFSET_ORIG0;
  OffsetOrigin[1] = OFFSET_ORIG1;
  OffsetOrigin[2] = OFFSET_ORIG2;
  VectorAdd(GalCenter, OffsetOrigin, GalCenter);       // Manual recentering.

  if(StepNum > NUMSTEP_RECENTER && StepNum % NUMSTEP_RECENTER == 0)
  {                              // Only from times to times.
    for(i = 0; i < NbPart; i++)  // And dont forget to offset all the 
    {                            // particles positions to this new origin.
      VectorSub(Part[i].pos,    GalCenter, Part[i].pos);     // PartPos = PartPos - GalCenter.
      VectorSub(Part[i].posold, GalCenter, Part[i].posold);  // PartPosold = PartPosold - GalCenter.
      VectorSub(Part[i].posnew, GalCenter, Part[i].posnew);  // PartPosnew = PartPosnew - GalCenter.
    }
    VectorZero(GalCenter);
  }
}

/*********************** SPECIFIC STET PLUG TO HMI **************************/
void PrintParticlesInfo(FILE *pFile)
{
  double RayGrid = (LG_GRID/2)/KILPARSEC;
  fprintfPerso(pFile, " RayGridKpc=%f", RayGrid);                  
}
