/*   Calculation of gravitational potential.
*/

#include <stdlib.h> 
#include <string.h> 
#include <stdio.h>
#include <math.h>
#include "common.h"
#include "hmi.h"
#include "init.h"
#include "particles.h"
#include "stet.h"
#include "period.h"
#include "potential.h"
#include "fft.h"

/* ----------------------------- The GRID ---------------------------------- */
#if NB_DIM == 2
  unsigned char GridDisk[2*NBCELLX_GRID][2*NBCELLY_GRID][NBCELLZ_GRID];
                                                        // NBCELLZ_GRID=1 here.
#elif NB_DIM == 3
  unsigned char GridDisk[2*NBCELLX_GRID][2*NBCELLY_GRID][2*NBCELLZ_GRID];
          // 2* for allowing the test of x, y, z out of bounds of regular grid.
#endif

/* ---------------------------- Disk array INIT ---------------------------- */
double InvDist[NBCELLX_FFT][NBCELLY_FFT][NBCELLZ_FFT];            // For perfs.
double Potential[NBCELLX_FFT][NBCELLY_FFT][NBCELLZ_FFT];   // Gravit potential.

/* -------------------- Optimisations of perfs only ------------------------ */
unsigned char GridOptimDisk[NBCELLX_GRID][NBCELLY_GRID][NBCELLZ_GRID];
unsigned char GridOptimSquare[NBCELLX_GRID][NBCELLY_GRID][NBCELLZ_GRID];

extern void InitDisks();          // for init below.
extern void InitInvDist();        // for init below.

/* --------------------------- Init of the grid ---------------------------- */
void InitPotential()
{
  memset(Potential, 0, sizeof(Potential));
  InitInvDist();
  if(MODE_CALCUL_POTENTIAL == MODE_FFT) InitFFT();
  else InitDisks();
}

/* ---------------------------- Disk array INIT ---------------------------- */
void InitDisks()
{
  unsigned int jx, jy, jz;
  
  memset(GridDisk, 0, sizeof(GridDisk));
  memset(GridOptimDisk, 0, sizeof(GridOptimDisk));
  memset(GridOptimSquare, 0, sizeof(GridOptimSquare));
  for(jy=0 ; jy<NBCELLY_GRID ; jy++)
  {
    for(jx=0 ; jx<NBCELLX_GRID ; jx++)
    {
      for(jz=0 ; jz<NBCELLZ_GRID ; jz++)
      {
        long int x = jx - GRIDDISK_RAY;
        long int y = jy - GRIDDISK_RAY;
        #if (NB_DIM == 3)
          long int z = jz - GRIDDISK_RAY;
        #else
          long int z = 0;
        #endif
			  if(x*x + y*y + z*z  <=  GRIDDISK_RAY * GRIDDISK_RAY)
			  {
          GridDisk[jx][jy][jz] = 1;               // This is inside the disk.
        }
        else ;                                    // Avoid square corners.
        if(jx*jx + jy*jy + jz*jz  <=  GRIDOPTIMDISK_RAY * GRIDOPTIMDISK_RAY)
        {   
          GridOptimDisk[jx][jy][jz] = 1;          // This is inside the disk.
        }
        else ;                                    // Avoid square corners.
        if(jx <= GRIDOPTIMSQUARE_RAY  &&  jy <= GRIDOPTIMSQUARE_RAY  &&  jz <= GRIDOPTIMSQUARE_RAY)
        {   
          GridOptimSquare[jx][jy][jz] = 1;        // This is inside the square.
        }
        else ;                                
      }
    }
  }
}

/* --------------------- InverseDistance array INIT -------------------- */
void InitInvDist()
{
  unsigned int jx, jy, jz;
  long double dx0, dx1, dx2, dx, distance;
  
  memset(InvDist, 0, sizeof(InvDist));

  /* --- Calculate inverse distance in small grid ---------------------- */
  for(jy=0 ; jy<=NBCELLY_GRID ; jy++)
  {
    for(jx=0 ; jx<=NBCELLX_GRID ; jx++)
    {
      for(jz=0 ; jz<=NBCELLZ_GRID ; jz++)
      {
        dx0 = jx * LGX_CELL;
        dx1 = jy * LGY_CELL;
        dx2 = jz * LGZ_CELL;
        dx = dx0*dx0 + dx1*dx1 + dx2*dx2;
        distance = sqrt(dx);
        if(jx!=0 || jy!=0 || jz!=0)              // Avoid division by zero.
        {
          InvDist[jx][jy][jz] = G_CONST / distance;  // Small grid (first quarter))
        }else ;
      }
    }
  }
  /* --- First corner of small and also large grid --------------------- */
  distance = LGX_CELL;
  InvDist[0][0][0] = G_CONST / distance;
}

#if GRAVIT_MODEL==STET
void GetAccelFromGalaxyStet(int PartInd, double accel[], unsigned int NumCell[])
{
	double StetCorr;        // Corrective term added to Newton's acceleration.

  /* --------------------- Calculate corrective term ---------------------- */
  StetCorr = GetStetCorrOut(PartInd, NumCell);

  /* ---------------- Muliply accelerations by corrective term ------------ */
	accel[0] *= StetCorr;
	accel[1] *= StetCorr;
  #if (NB_DIM == 3)
	  accel[2] *= StetCorr;
  #endif
}

#else  // NEWTON case (for linking only)
  void GetAccelFromGalaxyStet(int PartInd, double accel[], unsigned int NumCell[]){}
#endif

void GravitFieldCalculate()    // This function is only used at init time 
{                              // for some cases. No need for periodic mode.
  unsigned int PartInd;
  unsigned CR;
  unsigned int NumCell[3];
  #if (GRAVIT_MODEL == STET)
    double StetCorr;        // Corrective term added to Newton's acceleration.
    StetCorr = GetStetCorr((unsigned int) StetNbPartInit,0,0,0);
  #endif

  for(PartInd = 0 ; PartInd < NbPart ; PartInd++)
  {
    CR = GetCell(Part[PartInd].pos, NumCell);
    if(CR != CELL_IN) Error(200);
    
    /**** Calculate acceleration from previously calculated potential *****/
    #if (NB_DIM == 3)
      if(     NumCell[0]<=0  ||  NumCell[0]>=NBCELLX_GRID-1
          ||  NumCell[1]<=0  ||  NumCell[1]>=NBCELLY_GRID-1
          ||  NumCell[2]<=0  ||  NumCell[2]>=NBCELLZ_GRID-1
        )
    #else
      if(     NumCell[0]<=0  ||  NumCell[0]>=NBCELLX_GRID-1
          ||  NumCell[1]<=0  ||  NumCell[1]>=NBCELLY_GRID-1
        )
    #endif
    {    // On the fringe. Impossible to calculate acceleration in a coherent way.
      Part[PartInd].acc[0] = Part[PartInd].acc[1] = 0;
      #if (NB_DIM == 3)
        Part[PartInd].acc[2] = 0;
      #endif
    }
    else
    {                                /* Done in a "time reversal" manner */
        Part[PartInd].acc[0] =  Potential[NumCell[0]+1][NumCell[1]][NumCell[2]];
        Part[PartInd].acc[0] -= Potential[NumCell[0]-1][NumCell[1]][NumCell[2]];
        Part[PartInd].acc[0] /= 2 * LGX_CELL;
        Part[PartInd].acc[1] =  Potential[NumCell[0]][NumCell[1]+1][NumCell[2]];
        Part[PartInd].acc[1] -= Potential[NumCell[0]][NumCell[1]-1][NumCell[2]];
        Part[PartInd].acc[1] /= 2 * LGY_CELL;
      #if (NB_DIM == 3)
        Part[PartInd].acc[2] =  Potential[NumCell[0]][NumCell[1]][NumCell[2]+1];
        Part[PartInd].acc[2] -= Potential[NumCell[0]][NumCell[1]][NumCell[2]-1];
        Part[PartInd].acc[2] /= 2 * LGZ_CELL;
      #endif

      #if (GRAVIT_MODEL == STET)
        Part[PartInd].acc[0] *= StetCorr;
        Part[PartInd].acc[1] *= StetCorr;
        #if (NB_DIM == 3)
          Part[PartInd].acc[2] *= StetCorr;
        #endif
      #endif
    }
  }                 // End of for loop over the i particles.
}

// Tested OK with simul of 3D period Newton the 24 sept 2014.
void CalculatePotentialCell(unsigned int ix,unsigned int iy,unsigned int iz)
{
	int x, y, z, jx, jy, jz;
  unsigned int ajx, ajy, ajz;

  #if GRAVIT_MODEL == STET
    unsigned int NbPartInStet;
    double StetCorr;            // Corrective term added to Newton's accelerat.
    NbPartInStet = 0;
  #endif

	Potential[ix][iy][iz] = 0;

	for(jx=-NBCELLX_GRID/2 ; jx<NBCELLX_GRID/2 ; jx++)
	{
		for(jy=-NBCELLY_GRID/2 ; jy<NBCELLY_GRID/2 ; jy++)
		{
      #if (NB_DIM == 3)
        for(jz=-NBCELLZ_GRID/2 ; jz<NBCELLZ_GRID/2 ; jz++)
      #else
        jz = 0;
      #endif
        {
          x = GetModuloPeriodic(ix + jx, NBCELLX_GRID);
          y = GetModuloPeriodic(iy + jy, NBCELLY_GRID);
          z = GetModuloPeriodic(iz + jz, NBCELLZ_GRID);
          ajx = abs(jx); ajy = abs(jy); ajz = abs(jz);  
          if(IsValidCell(x,y,z) == 1)  
          {                    // Only for cell inside the disk.
            double CellMass = CellNbPart[x][y][z] * PartMass;
            Potential[ix][iy][iz] += CellMass * InvDist[ajx][ajy][ajz];
          }
          else continue;      // Avoid : outside the disk.  

          #if GRAVIT_MODEL == STET
            if(InsideStetDisk(ix, iy, iz, ajx, ajy, ajz))
            {                    // Only for cell inside Stet disk.
              NbPartInStet += CellNbPart[x][y][z];
            }        // At the end nb of all particles in Stet disk is memorized.
            else continue;      // Avoid : outside Stet disk.  
          #endif
        }
	  	}
	}
  #if GRAVIT_MODEL == STET
    StetCorr = GetStetCorr(NbPartInStet,ix,iy,iz);
    StetCorrMemo[ix][iy][iz] = StetCorr;           // DEBUG ONLY !
    StetCorrMax = max(StetCorr, StetCorrMax);
    StetCorrMin = min(StetCorr, StetCorrMin);
    Potential[ix][iy][iz] *= StetCorr;
    NbPartInStetDebug[ix][iy][iz] = NbPartInStet;   // DEBUG ONLY !
  #endif
}

void CalculatePotentialCell3(unsigned int ix, unsigned int iy, unsigned int iz)
{
  double CellMass;
	int jx, jy, jz, x, y, z;
  unsigned int ajx, ajy, ajz;
  unsigned long int NbP;

  #if GRAVIT_MODEL == STET
    unsigned int NbPartInStet;
    double StetCorr;        // Corrective term added to Newton's acceleration.
  #endif

	Potential[ix][iy][iz] = 0;             // And it will be added to this below.

  #if GRAVIT_MODEL == STET
  	NbPartInStet = 0;
  #endif

	/************* Potential due to very close small cells *********************/
	for(jx = -GRIDOPTIMSQUARE_RAY ; jx <= GRIDOPTIMSQUARE_RAY ; jx++)
	{                                // Inside of the very close optimizing square.
		for(jy = -GRIDOPTIMSQUARE_RAY ; jy <= GRIDOPTIMSQUARE_RAY ; jy++)
		{
      #if (NB_DIM == 3)
        for(jz = -GRIDOPTIMSQUARE_RAY ; jz <= GRIDOPTIMSQUARE_RAY ; jz++)
        {
      #else
         jz = 0;
      #endif
          x = GetModuloPeriodic(ix + jx, NBCELLX_GRID);
          y = GetModuloPeriodic(iy + jy, NBCELLY_GRID);
          z = GetModuloPeriodic(iz + jz, NBCELLZ_GRID);      
          if(IsValidCell(x,y,z))    
          {                         // If this cell is inside the disk.
            CellMass = CellNbPart[x][y][z] * PartMass;
            ajx = abs(jx); ajy = abs(jy); ajz = abs(jz);
            Potential[ix][iy][iz] += CellMass * InvDist[ajx][ajy][ajz];

            #if GRAVIT_MODEL == STET
              if(InsideStetDisk(ix, iy, iz, ajx, ajy, ajz))
              {                    // Only for cell inside Stet disk.
                NbPartInStet += CellNbPart[x][y][z];
              }                 // At end nb of partic in Stetdisk is memorized.
              else continue;    // Avoid : outside Stet disk.  
            #endif
          }
          else ;                   // Avoid if out of GridDisk.
      #if (NB_DIM == 3)
        }
      #endif
		}
	}

	for(jx = -MAXIND_BIGCELLX ; jx <= MAXIND_BIGCELLX ; jx+=NBCELL_OPTIM_INCR)
	{
		for(jy = -MAXIND_BIGCELLX ; jy <= MAXIND_BIGCELLX ; jy+=NBCELL_OPTIM_INCR)
		{                           // for jx=jy=0, Potential Potential[jx][jy] = 0
      #if (NB_DIM == 3)
        for(jz = -MAXIND_BIGCELLX ; jz <= MAXIND_BIGCELLX ; jz+=NBCELL_OPTIM_INCR)
        { 
      #else
          jz = 0;
      #endif
          x = GetModuloPeriodic(ix+jx, NBCELLX_GRID);
          y = GetModuloPeriodic(iy+jy, NBCELLY_GRID);
          z = GetModuloPeriodic(iz+jz, NBCELLZ_GRID);
          ajx = abs(jx); ajy = abs(jy); ajz = abs(jz);

          if(IsValidBigCell(x,y,z))    
          {              // If the center of this large cell is inside the disk.
            /* ------------- Large cell FAR : optimisation ----------------- */
            if(GridOptimDisk[ajx][ajy][ajz] == 0)
            {                                 // Out of optimizing disk : jx+jy too strong.
              NbP = CellNbPart[x][y][z] * NBSMALLCELL_IN_LARGECELL; 
                                                          // Only centered small cell is read 
            }                                             // for this large cell.

            /* ------------- Large cell close : calculate normally --------- */
            else if(GridOptimSquare[ajx][ajy][ajz] == 0)  // If outside of very close square.
            {                                             // Large cell close to us.
              NbP =   CellNbPart[x-1][y-1][z]             //  Calculate each small cell.
                    + CellNbPart[x-1][y]  [z]
                    + CellNbPart[x-1][y+1][z]
                    + CellNbPart[x]  [y-1][z]
                    + CellNbPart[x]  [y]  [z]
                    + CellNbPart[x]  [y+1][z]
                    + CellNbPart[x+1][y-1][z]
                    + CellNbPart[x+1][y]  [z]
                    + CellNbPart[x+1][y+1][z];
              #if (NB_DIM == 3)
                NbP +=  CellNbPart[x-1][y-1][z-1]
                      + CellNbPart[x-1][y]  [z-1]
                      + CellNbPart[x-1][y+1][z-1]
                      + CellNbPart[x]  [y-1][z-1]
                      + CellNbPart[x]  [y]  [z-1]
                      + CellNbPart[x]  [y+1][z-1]
                      + CellNbPart[x+1][y-1][z-1]
                      + CellNbPart[x+1][y]  [z-1]
                      + CellNbPart[x+1][y+1][z-1];
                NbP +=  CellNbPart[x-1][y-1][z+1]
                      + CellNbPart[x-1][y]  [z+1]
                      + CellNbPart[x-1][y+1][z+1]
                      + CellNbPart[x]  [y-1][z+1]
                      + CellNbPart[x]  [y]  [z+1]
                      + CellNbPart[x]  [y+1][z+1]
                      + CellNbPart[x+1][y-1][z+1]
                      + CellNbPart[x+1][y]  [z+1]
                      + CellNbPart[x+1][y+1][z+1];
              #endif
            }
            else
            {         // Large cell close to us. Has been calc in small loop above.
              continue;
            }
            #if GRAVIT_MODEL == STET
              if(InsideStetDisk(ix, iy, iz, ajx, ajy, ajz))
              {                    // Only for cell inside Stet disk.
                NbPartInStet += NbP;
              }                    // At the end nb of all particles in Stet disk is memorized.
              else ;               // Avoid : outside Stet disk.  
            #endif

            CellMass = NbP * PartMass;
            Potential[ix][iy][iz] += CellMass * InvDist[ajx][ajy][ajz];
          }
          else
          {
            continue;              // Avoid : outside the disk.  
          } 
      #if (NB_DIM == 3)
        }
      #endif
		}
	}

  #if GRAVIT_MODEL == STET
    StetCorr = GetStetCorr(NbPartInStet,ix,iy,iz);
    StetCorrMax = max(StetCorr, StetCorrMax);
    StetCorrMin = min(StetCorr, StetCorrMin);
    Potential[ix][iy][iz] *= StetCorr;
    NbPartInStetDebug[ix][iy][iz] = NbPartInStet;   // For DEBUG ONLY
    StetCorrMemo[ix][iy][iz] = StetCorr;            // For DEBUG ONLY
  #endif
}
